package voidious;

import robocode.*;

import java.awt.Color;
import java.awt.Graphics2D;
import voidious.radar.DiamondEyes;
import voidious.gun.DiamondFist;
import voidious.move.DiamondWhoosh;

/**
 * Diamond - a robot by Voidious
 *
 * A melee specialist. Employs Minimum Risk movement in general and Wave
 * Surfing (derived from Dookious) in 1v1 scenarios. 
 * 
 * Uses k-nearest neighbors and kernel density (aka "Dynamic Clustering")
 * for surfing and targeting data analysis. Uses displacement vectors for 
 * recording / reconstructing firing angles in the gun, GuessFactors in the
 * surfing movement.
 *
 * Code is open source, released under the RoboWiki Public Code License:
 * http://robowiki.net/cgi-bin/robowiki?RWPCL
 */

public class Diamond extends AdvancedRobot {
    protected static final boolean _TC = false;
    protected static final boolean _MC = true;
    
    protected static DiamondEyes _radar;
    protected static DiamondWhoosh _move;
    protected static DiamondFist _gun;
    
    protected static long _lastPaintRound = 0;
    protected static long _lastPaintTime = -2;
    
    public void run() {
        if (_radar == null) { _radar = new DiamondEyes(this); }
        if (_move == null) { _move = new DiamondWhoosh(this); }
        if (_gun == null) { _gun = new DiamondFist(this, _radar, _TC); }
        
        setAdjustGunForRobotTurn(true);
        setAdjustRadarForGunTurn(true);
        setColors(Color.black, Color.black, new Color(255, 255, 170));
        
        _radar.initRound(this);
        _move.initRound(this);
        _gun.initRound(this);
        
        while (true) {
            updatePaintProcessing();
            if (!_TC) {
                _move.execute();
            }
            if (!_MC) {
                _gun.execute();
            }
            _radar.execute();
            execute();
        }
    }
    
    public void onScannedRobot(ScannedRobotEvent e) {
        _radar.onScannedRobot(e);
        if (!_TC) {
            _move.onScannedRobot(e);
        }
        if (!_MC) {
            _gun.onScannedRobot(e);
        }
    }
    
    public void onRobotDeath(RobotDeathEvent e) {
        _radar.onRobotDeath(e);
        if (!_TC) {
            _move.onRobotDeath(e);
        }
        if (!_MC) {
            _gun.onRobotDeath(e);
        }
    }
    
    public void onHitByBullet(HitByBulletEvent e) {
        if (!_TC) {
            _move.onHitByBullet(e);
        }
        if (!_MC) {
            _gun.onHitByBullet(e);
        }
    }
    
    public void onBulletHit(BulletHitEvent e) {
        if (!_TC) {
            _move.onBulletHit(e);
        }
        if (!_MC) {
            _gun.onBulletHit(e);
        }
    }
    
    public void onBulletHitBullet(BulletHitBulletEvent e) {
        if (!_TC) {
            _move.onBulletHitBullet(e);
        }
    }
    
    public void onHitWall(HitWallEvent e) {
        System.out.println("WARNING: I hit a wall.");
    }
    
    public void onWin(WinEvent e) {
        if (!_MC) {
            _gun.onWin(e);
        }
        if (!_TC) {
            _move.onWin(e);
        }
    }

    public void onDeath(DeathEvent e) {
        if (!_MC) {
            _gun.onDeath(e);
        }
        if (!_TC) {
            _move.onDeath(e);
        }
    }
    
    public void onPaint(Graphics2D g) {
        _lastPaintRound = getRoundNum();
        _lastPaintTime = getTime();
        
        _radar.onPaint(g);
        if (!_TC) {
            _move.onPaint(g);
        }
        if (!_MC) {
            _gun.onPaint(g);
        }
    }

    public void onSkippedTurn(SkippedTurnEvent e) {
        System.out.println("WARNING: Turn skipped at: " + e.getTime());
    }

    protected void updatePaintProcessing() {
        if (_lastPaintRound == getRoundNum() && 
                getTime() - _lastPaintTime <=1) {
            _move.paintOn();
            _gun.paintOn();
        } else {
            _move.paintOff();
            _gun.paintOff();            
        }
    }
}
