package voidious.gun;

import robocode.*;
import robocode.util.Utils;

import java.util.HashMap;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Vector;
import java.util.Arrays;
import java.awt.Graphics2D;
import java.awt.geom.Point2D;
import java.awt.Color;
import java.awt.geom.Rectangle2D;

import voidious.radar.DiamondEyes;
import voidious.gun.EnemyDataGun;
import voidious.utils.RoboGraphic;
import voidious.utils.DiaWave;
import voidious.utils.DiaUtils;
import voidious.utils.KdBucketTree;

/**
 * Code is open source, released under the RoboWiki Public Code License:
 * http://robowiki.net/?RWPCL
 */

public class DiamondFist {
    protected static final boolean ENABLE_DEBUGGING_GRAPHICS = true;
    protected static final double CLUSTER_SIZE = 30;
    protected static final double CURRENT_TARGET_BIAS = 0.8;
    
    protected static final double LONGER_THAN_LONGEST_DISTANCE = 50000;
    protected static final double DEFAULT_ENERGY = 100;
    protected static final double BOT_WIDTH = 36;
    protected static final double BOT_HALF_WIDTH = 18;
    protected static final String NO_TARGET = "";
    protected static final double NO_FIRING_ANGLE = -999;
    protected static final boolean WAVE_AIMING = true;
    protected static final boolean WAVE_COLLECTING = false;
    
    protected AdvancedRobot _robot;
    protected DiamondEyes _radar; 
    protected boolean _tcMode;
    protected HashMap<String, EnemyDataGun> _enemies;
    protected ArrayList<DiaWave> _waves;
    protected Point2D.Double _myLocation;
    protected String _targetName;
    protected double _targetRating;
    protected double _bulletPower;
    protected int _enemiesAlive;
    protected int _enemiesTotal;
    protected Rectangle2D.Double _fieldRect;
    protected double _battleFieldWidth;
    protected double _battleFieldHeight;
    protected Point2D.Double _centerField;
    
    protected double _damageGiven;
    protected Vector<RoboGraphic> _renderables;
    protected boolean _painting;
    
    public DiamondFist(AdvancedRobot robot, DiamondEyes radar,
        boolean isTc) {
        _robot = robot;
        _radar = radar;
        _tcMode = isTc;
        _enemies = new HashMap<String, EnemyDataGun>();
        _waves = new ArrayList<DiaWave>();
        _enemiesTotal = _robot.getOthers();
        _battleFieldWidth = _robot.getBattleFieldWidth();
        _battleFieldHeight = _robot.getBattleFieldHeight();
        _fieldRect = new java.awt.geom.Rectangle2D.Double(18, 18, 
                _battleFieldWidth - 36, _battleFieldHeight - 36);
        _centerField = new Point2D.Double(_robot.getBattleFieldWidth() / 2,
            _robot.getBattleFieldHeight() / 2);
        
        _damageGiven = 0;
        _renderables = new Vector<RoboGraphic>();
        _painting = false;
    }
    
    public void initRound(AdvancedRobot robot) {
        _robot = robot;
        Iterator<EnemyDataGun> edgIterator = _enemies.values().iterator();
        while (edgIterator.hasNext()) {
            EnemyDataGun enemyData = edgIterator.next();
            enemyData.energy = DEFAULT_ENERGY;
            enemyData.distance = LONGER_THAN_LONGEST_DISTANCE;
            enemyData.alive = true;
            enemyData.clearDistancesSq();
            enemyData.pastLocations.clear();
        }
        _waves.clear();
        _targetName = NO_TARGET;
        _targetRating = Double.POSITIVE_INFINITY;
        _myLocation = new Point2D.Double(_robot.getX(), _robot.getY());
        _bulletPower = 3;
        _enemiesAlive = _robot.getOthers();
        
        _renderables.clear();
    }
    
    public void execute() {
        _myLocation = new Point2D.Double(_robot.getX(), _robot.getY());
        _enemiesAlive = _robot.getOthers();
        long currentTime = _robot.getTime();

        Iterator<EnemyDataGun> edgIterator = _enemies.values().iterator();
        while (edgIterator.hasNext()) {
            EnemyDataGun enemyData = edgIterator.next();
            if (enemyData.lastScanTime < currentTime) {
                enemyData.pastLocations.add(enemyData.location);
            }
        }

        updateBotDistancesSq();
        evalBulletPower();
        checkActiveWaves();
        aimAndFire();
        _myLocation = DiaUtils.nextLocation(_robot);
    }

    public void aimAndFire() {
        if (_targetName != NO_TARGET) {
            EnemyDataGun enemyData = _enemies.get(_targetName);
            DiaWave fireWave = enemyData.lastWaveFired;
            Point2D.Double myNextLocation = DiaUtils.nextLocation(_robot);

            fireWave.setBulletPower(_bulletPower);

            // Leaves at least 2 ticks for aiming. (An extra one because the
            // difference could still get down to 2 in this setup, radar
            // turning one tick further away as one tick also goes by.)
            if (ticksUntilGunCool() - _radar.minTicksToScan(_targetName) <= 3) {
                _radar.setRadarLock(_targetName);
            } else {
                _radar.releaseRadarLock();
            }
            
            double firingAngle;
            if (enemyData.energy == 0 ||
                _robot.getGunHeat() / _robot.getGunCoolingRate() > 3) {
                firingAngle = DiaUtils.absoluteBearing(
                    myNextLocation, fireWave.targetLocation);
            } else {
                firingAngle = aimWithWave(fireWave);
            }
            
            _robot.setTurnGunRightRadians(Utils.normalRelativeAngle(
                firingAngle - _robot.getGunHeadingRadians()));

            if (Math.abs(_robot.getGunTurnRemainingRadians())
                    < DiaUtils.botWidthAimAngle(
                        myNextLocation.distance(fireWave.targetLocation)) || 
                    _enemiesAlive > 3) {
                if (_tcMode) {
                    _robot.setFire(3);
                } else if (_robot.getEnergy() > _bulletPower) {
                    _robot.setFire(_bulletPower);
                }
            }
        }
    }
/*
    public void aimAndFireHot() {
        if (_targetName != NO_TARGET) {
            if (ticksUntilGunCool() - _radar.minTicksToScan(_targetName) <= 3) {
                _radar.setRadarLock(_targetName);
            } else {
                _radar.releaseRadarLock();
            }
            
            double bearingToTarget = DiaUtils.absoluteBearing(_myLocation, 
                _enemies.get(_targetName).location);
            _robot.setTurnGunRightRadians(Utils.normalRelativeAngle(
                bearingToTarget - _robot.getGunHeadingRadians()));

            if ((Math.abs(_robot.getGunTurnRemainingRadians())
                < DiaUtils.botWidthAimAngle(_enemies.get(_targetName).distance) 
                || _enemiesAlive > 3) && _robot.getEnergy() > 0.3) {

                if (_robot.getEnergy() <= 3) {
                    _robot.setFire(0.3);
                } else {
                    _robot.setFire(3);
                }
            }
        }
    }
*/
    public void evalBulletPower() {
        EnemyDataGun enemyData = _enemies.get(_targetName);
        if (enemyData == null) { return; }
        
        if (_tcMode) {
            _bulletPower = Math.min(_robot.getEnergy(), 3);
        } else { 
            _bulletPower = 2;
            
            if (enemyData.distance < 300 || _enemiesAlive >= 6) {
                _bulletPower = 3;
            }
            
            if (enemyData.distance > 700 && (_enemiesAlive <= 3)) {
                _bulletPower = 1.4;
            }
            
            if (_robot.getEnergy() < 25 && _enemiesAlive <= 3) {
                _bulletPower = Math.min(_bulletPower, 2 - 
                    ((25 - _robot.getEnergy()) / 12));
            }
        }
    }
   
    public void onScannedRobot(ScannedRobotEvent e) {
        String botName = e.getName();
        Point2D.Double enemyLocation = DiaUtils.project(_myLocation, 
            Utils.normalAbsoluteAngle(
                e.getBearingRadians() + _robot.getHeadingRadians()), 
            e.getDistance());
        
        EnemyDataGun enemyData;
        if (_enemies.containsKey(botName)) {
            enemyData = _enemies.get(botName);
            enemyData.timeSinceDirectionChange += 
                e.getTime() - enemyData.lastScanTime;
            enemyData.timeSinceVelocityChange += 
                e.getTime() - enemyData.lastScanTime;

            if (Math.abs(e.getVelocity() - 
                    enemyData.lastNonZeroVelocity) > 0.5) {
                enemyData.timeSinceVelocityChange = 0;
            }
            
            if (Math.abs(e.getVelocity()) > 0.5) {
                if (DiaUtils.nonZeroSign(e.getVelocity()) != 
                    DiaUtils.nonZeroSign(enemyData.lastNonZeroVelocity)) {
                    enemyData.timeSinceDirectionChange = 0;
                }
                enemyData.lastNonZeroVelocity = e.getVelocity();
            }

            enemyData.distance = e.getDistance();
            enemyData.energy = e.getEnergy();
            enemyData.location = enemyLocation;
            enemyData.lastScanTime = e.getTime();
            enemyData.previousVelocity = enemyData.velocity;
            enemyData.velocity = e.getVelocity();
            enemyData.heading = e.getHeadingRadians();
        } else {
            enemyData = new EnemyDataGun(e.getDistance(), e.getEnergy(), 
                enemyLocation, e.getTime(), e.getVelocity(), 
                e.getHeadingRadians());
            _enemies.put(botName, enemyData);
        }
        enemyData.pastLocations.add(enemyLocation);

        fireNextTickWave(enemyLocation, botName, _bulletPower);

        double botTargetEval = Math.max(10, enemyData.energy) * 
            DiaUtils.square(enemyData.distance) *
            (15 + Math.sqrt(cornerDistance(enemyData.location)));
//            (_enemiesAlive <= 4 ? 1 :
//                (15 + Math.sqrt(cornerDistance(enemyData.location))));
//            (10 + (enemyData.damageTaken / (_robot.getRoundNum() + 1)));
        if (enemyData.alive && (_targetName.equals(botName) || is1v1() ||
            _targetName.equals(NO_TARGET) ||
            (botTargetEval < _targetRating * CURRENT_TARGET_BIAS && 
                // Leaves at least 2 ticks for aiming.
                ticksUntilGunCool() - _radar.minTicksToScan(botName) >= 2))) {
            _targetName = botName;
            _targetRating = botTargetEval;
        }
        
    }
    
    public void onRobotDeath(RobotDeathEvent e) {
        String botName = e.getName();

        try {
            _enemies.get(botName).alive = false;
            if (botName.equals(_targetName)) {
                _targetRating = Double.POSITIVE_INFINITY;
                _targetName = NO_TARGET;
                _radar.releaseRadarLock();
            }
        } catch (java.lang.NullPointerException npe) {
            System.out.println(
                "WARNING (gun):  A bot died that I never knew existed!");
        }
    }
    
    public void onHitByBullet(HitByBulletEvent e) {
        String botName = e.getName();
        
        try {
           _enemies.get(botName).damageTaken += 
               Rules.getBulletDamage(e.getBullet().getPower());
        } catch (java.lang.NullPointerException npe) {
            System.out.println(
                "WARNING (gun):  A bot shot me that I never knew existed!");
        }
    }

    public void onWin(WinEvent e) {
        roundOver();
    }

    public void onDeath(DeathEvent e) {
        roundOver();
    }
    
    public void onBulletHit(BulletHitEvent e) {
        _damageGiven += Math.min(
            Rules.getBulletDamage(e.getBullet().getPower()),
            _enemies.get(e.getName()).energy);
        
        String botName = e.getName();
        
        try {
           _enemies.get(botName).damageGiven += 
               Rules.getBulletDamage(e.getBullet().getPower());
        } catch (java.lang.NullPointerException npe) {
            System.out.println(
                "WARNING (gun):  I shot a bot that I never knew existed!");
        }
    }

    public void roundOver() {
        if (_tcMode) {
            System.out.println("TC score: " +
                _damageGiven / (_robot.getRoundNum() + 1));
        }
    }

    public void onPaint(Graphics2D g) {
        if (_painting) {
            Iterator<RoboGraphic> i = _renderables.iterator();
            while(i.hasNext()){
                RoboGraphic r = i.next();
                r.render(g);
            }
            _renderables.clear();
        }
    }
    
    public DiaWave fireNextTickWave(Point2D.Double targetLocation, 
        String targetName, double bulletPower) {

        EnemyDataGun targetData = _enemies.get(targetName);
        Point2D.Double myNextLocation = DiaUtils.nextLocation(_robot);
        Point2D.Double enemyNextLocation = DiaUtils.nextLocation(targetLocation, 
            targetData.heading, targetData.velocity);
        Point2D.Double orbitLocation = closestBotToEnemy(targetName);
        
        double accel = DiaUtils.limit(-Rules.DECELERATION, 
                DiaUtils.accel(targetData.velocity, targetData.previousVelocity), 
                Rules.ACCELERATION);
/*
        if (_painting) {
            _renderables.add(RoboGraphic.drawLine(
                enemyNextLocation, orbitLocation, Color.green));
            _renderables.add(RoboGraphic.drawCircle(
                enemyNextLocation, 35, Color.green));
            _renderables.add(RoboGraphic.drawCircle(
                orbitLocation, 30, Color.gray));
            double angleToOrbit = DiaUtils.absoluteBearing(enemyNextLocation, 
                    orbitLocation);
            _renderables.addAll(Arrays.asList(
                RoboGraphic.drawArrowHead(
                    DiaUtils.project(enemyNextLocation, angleToOrbit,
                        enemyNextLocation.distance(orbitLocation) / 2 + 20),
                        30, angleToOrbit, Color.green)));
        }
*/
        DiaWave nextWave = new DiaWave(myNextLocation, enemyNextLocation, 
            orbitLocation, _robot.getTime() + 1, bulletPower, targetName, 
            targetData.heading, targetData.velocity, accel, 
            DiaUtils.nonZeroSign(targetData.lastNonZeroVelocity),
            orbitLocation.distance(enemyNextLocation), 
            targetData.timeSinceDirectionChange, 
            targetData.timeSinceVelocityChange, 0, 0,
            targetLocation.distance(targetData.getPastLocation(
                Math.min(8, targetData.pastLocations.size()-1))),
            targetLocation.distance(targetData.getPastLocation(
                Math.min(20, targetData.pastLocations.size()-1))),
            targetLocation.distance(targetData.getPastLocation(
                Math.min(40, targetData.pastLocations.size()-1))),
            _enemiesAlive, _robot.getGunHeat(), _fieldRect,
            _battleFieldWidth, _battleFieldHeight);

        // A bit convoluted, but it's easier and faster to execute this way.
        nextWave.targetWallDistance = is1v1() ?
            nextWave.preciseEscapeAngle(DiaWave.POSITIVE_GUESSFACTOR) /
                Math.asin(8.0 / nextWave.bulletSpeed) :
            Math.min(1.5, DiaUtils.directToWallDistance(enemyNextLocation, 
                targetData.distance, targetData.heading + 
                    (targetData.lastNonZeroVelocity > 0 ? 0 : Math.PI),
                bulletPower, _fieldRect));

        nextWave.targetRevWallDistance = is1v1() ?
            nextWave.preciseEscapeAngle(DiaWave.NEGATIVE_GUESSFACTOR) /
                Math.asin(8.0 / nextWave.bulletSpeed) :
             Math.min(1.5, DiaUtils.directToWallDistance(enemyNextLocation, 
                 targetData.distance, targetData.heading + 
                     (targetData.lastNonZeroVelocity > 0 ? Math.PI : 0),
                 bulletPower, _fieldRect));
        
        targetData.lastWaveFired = nextWave;
        if (_robot.getEnergy() != 0) {
            _waves.add(nextWave);
        }

        return nextWave;
    }
    
    public void checkActiveWaves() {
        long currentTime = _robot.getTime();

        Iterator<DiaWave> wavesIterator = _waves.iterator();
        while (wavesIterator.hasNext()) {
            DiaWave w = wavesIterator.next();
            EnemyDataGun edg = _enemies.get(w.botName);

            if (w.fireTime == currentTime) {
                w.gunHeat = _robot.getGunHeat();
                boolean recalcWallDistance = false;
                if (w.bulletPower != _bulletPower) {
                    // If bullet power changes between segmentation tick
                    // and firing tick, update with correct power and
                    // related values (ie, wall distance).

                    w.setBulletPower(_bulletPower);
                    
                    recalcWallDistance = true;
                }
                
                if (edg.lastScanTime == w.fireTime) {
                    // If we get two scans in a row, we can overwrite our
                    // estimates with empirical data.
                    w.targetLocation = edg.location;
                    w.sourceLocation = _myLocation;
                    w.absBearing = DiaUtils.absoluteBearing(
                        w.sourceLocation, w.targetLocation);
                    
                    // Technically might recalc wall distance here, too,
                    // but lotsa CPU for negligible difference.
                }
                
                if (recalcWallDistance) {

                    // TODO: Lots of duplicated code, so ugly! Refactor.

                    w.clearCachedPreciseEscapeAngles();
                    w.targetWallDistance = is1v1() ?
                        w.preciseEscapeAngle(DiaWave.POSITIVE_GUESSFACTOR) /
                            Math.asin(8.0 / w.bulletSpeed) :
                        Math.min(1.5, DiaUtils.directToWallDistance(
                            w.targetLocation, w.targetDistance, 
                            w.effectiveHeading(), w.bulletPower, _fieldRect));

                    w.targetRevWallDistance = is1v1() ?
                        w.preciseEscapeAngle(DiaWave.NEGATIVE_GUESSFACTOR) /
                            Math.asin(8.0 / w.bulletSpeed) :
                        Math.min(1.5, DiaUtils.directToWallDistance(
                            w.targetLocation, w.targetDistance, 
                            w.effectiveHeading() +  Math.PI, w.bulletPower, 
                            _fieldRect));
                }
            }
            
            if (!edg.alive) {
                wavesIterator.remove();
            } else if (w.wavePassed(edg.location, edg.lastScanTime, currentTime)) {
                Point2D.Double dispVector = w.displacementVector();
                double[] dataPoint = dataPointFromWave(w, WAVE_COLLECTING);
                edg.scanTree.insert(dataPoint);
                edg.displacementVectors.put(dataPoint, dispVector);
            
                if (_painting) {
                    double drawBearing = (w.orbitDirection * 
                        DiaUtils.absoluteBearing(DiaWave.ORIGIN, dispVector)) +
                        w.effectiveHeading();
                    double drawDistance = dispVector.distance(DiaWave.ORIGIN) *
                        w.waveBreakBulletTicks();
                    _renderables.add(RoboGraphic.drawLine(
                        DiaUtils.project(w.targetLocation, w.targetHeading, 
                            50),
                        DiaUtils.project(w.targetLocation, 
                            w.targetHeading + Math.PI, 50), 
                        Color.darkGray));
                    _renderables.add(RoboGraphic.drawLine(
                        DiaUtils.project(w.targetLocation, 
                            w.targetHeading + (Math.PI / 2), 
                            50),
                        DiaUtils.project(w.targetLocation, 
                            w.targetHeading - (Math.PI / 2), 50), 
                        Color.darkGray));
                    _renderables.addAll(Arrays.asList(
                        RoboGraphic.drawArrowHead(DiaUtils.project(
                            w.targetLocation, w.effectiveHeading(),
                            50), 12, w.effectiveHeading(), Color.gray)));
                    _renderables.addAll(Arrays.asList(
                        RoboGraphic.drawArrowHead(DiaUtils.project(
                            w.targetLocation, 
                            w.effectiveHeading() + Math.PI,
                            50), 12, w.effectiveHeading(), Color.gray)));
                    _renderables.add(RoboGraphic.drawLine(w.targetLocation,
                        DiaUtils.project(w.targetLocation, drawBearing, 
                            drawDistance),
                        Color.red));
                    _renderables.addAll(Arrays.asList(
                        RoboGraphic.drawArrowHead(
                            DiaUtils.project(w.targetLocation, drawBearing, 
                                drawDistance / 2 + 20), 30, drawBearing, 
                            Color.red)));
                    _renderables.add(RoboGraphic.drawCircleFilled(
                        w.waveBreakLocation, Color.red, 4));
                    _renderables.add(RoboGraphic.drawCircle(
                        w.targetLocation, 25, Color.red));
                    _renderables.add(RoboGraphic.drawText(
                        "t-" + (currentTime-w.fireTime), 
                        w.targetLocation.x - 10, 
                        w.targetLocation.y - 5, Color.white));
                }
                wavesIterator.remove();
            }
        }
    }
    
    public double aimWithWave(DiaWave w) {
        EnemyDataGun edg = _enemies.get(w.botName);
        if (edg.displacementVectors.size() < 10) {
            return w.absBearing;
        }
        
        KdBucketTree scanTree = edg.scanTree;
        int clusterSize = (int)
            Math.min(edg.displacementVectors.size() / 10, CLUSTER_SIZE);
        double[] weights = is1v1() ? WEIGHTS_1V1 : WEIGHTS_FFA;
        double[] wavePoint = dataPointFromWave(w, WAVE_AIMING);
        double[][] nearestNeighbors = KdBucketTree.nearestNeighbors(
            scanTree, wavePoint, clusterSize, weights);
        int numScans = nearestNeighbors.length;
        if (numScans == 0) {
            return w.absBearing;
        }

        double[] firingAngles = new double[numScans];
        for (int x = 0; x < numScans; x++) {
            Point2D.Double dispVector = 
                edg.displacementVectors.get(nearestNeighbors[x]);
            Point2D.Double projectedLocation = w.projectLocation(dispVector);
            if (!_fieldRect.contains(projectedLocation)) {
                firingAngles[x] = NO_FIRING_ANGLE;
            } else {
                firingAngles[x] = Utils.normalRelativeAngle(
                    w.firingAngleFromTargetLocation(projectedLocation) 
                        - w.absBearing);
            }
        }
        
        double bestAngle = NO_FIRING_ANGLE;
        double bestDensity = Double.NEGATIVE_INFINITY;
        double bandwidth = DiaUtils.botWidthAimAngle(
            _myLocation.distance(w.targetLocation));
        
        for (int x = 0; x < numScans; x++) {
            if (firingAngles[x] == NO_FIRING_ANGLE) { continue; }

            double xFiringAngle = firingAngles[x];
            
            double xDensity = 0;
            for (int y = 0; y < numScans; y++) {
                if (x == y || firingAngles[y] == NO_FIRING_ANGLE) { continue; }
                
                double yFiringAngle = firingAngles[y];
                
                double ux = (xFiringAngle - yFiringAngle) / bandwidth;
                    
                // Gaussian
                  xDensity += Math.exp(-0.5 * ux * ux);
                
                // Triangle
//                if (Math.abs(ux) <= 1) {
//                    xDensity += 1 - Math.abs(ux);
//                }
            }
            
//            xDensity /= .000001 + 
//                KdBucketTree.manhattanDistance(wavePoint, nearestNeighbors[x], weights);

            if (xDensity > bestDensity) {
                bestAngle = xFiringAngle;
                bestDensity = xDensity;
            }
            
            if (_painting) {
                Point2D.Double angleHead = DiaUtils.project(w.sourceLocation, 
                    w.absBearing + xFiringAngle, Math.min(400, 
                        w.sourceLocation.distance(w.targetLocation) - 55));
                       
                _renderables.add(RoboGraphic.drawLine(w.sourceLocation,
                    angleHead, Color.red));
                _renderables.addAll(Arrays.asList(RoboGraphic.drawArrowHead(
                    angleHead, 10, w.absBearing + xFiringAngle, Color.red)));
            }
        }

        if (bestAngle == NO_FIRING_ANGLE) {
            return w.absBearing;
        }

        if (_painting) {
            double lineLength = Math.min(425, 
                    w.sourceLocation.distance(w.targetLocation) - 30);
            Point2D.Double angleHead = DiaUtils.project(w.sourceLocation, 
                w.absBearing + bestAngle, lineLength);
                   
            _renderables.add(RoboGraphic.drawLine(w.sourceLocation,
                angleHead, Color.yellow));
            _renderables.addAll(Arrays.asList(RoboGraphic.drawArrowHead(
                angleHead, 10, w.absBearing + bestAngle, Color.yellow)));
            _renderables.add(RoboGraphic.drawLine(
                DiaUtils.project(w.sourceLocation, w.absBearing + bestAngle - 
                    bandwidth, lineLength - 20),
                DiaUtils.project(w.sourceLocation, w.absBearing + bestAngle + 
                    bandwidth, lineLength - 20), 
                Color.yellow));
        }

        return Utils.normalAbsoluteAngle(w.absBearing + bestAngle);
    }
    
    protected long ticksUntilGunCool() {
        return Math.round(Math.ceil(
            _robot.getGunHeat() / _robot.getGunCoolingRate()));
    }

    protected void updateBotDistancesSq() {
        if (_enemies.size() <= 1) { return; }

        String[] botNames = new String[_enemies.size()];
        _enemies.keySet().toArray(botNames);
        
        for (int x = 0; x < botNames.length; x++) {
            EnemyDataGun edg1 = _enemies.get(botNames[x]);
            for (int y = x + 1; y < botNames.length; y++) {
                EnemyDataGun edg2 = _enemies.get(botNames[y]);
                if (edg1.alive && edg2.alive) {
                    double distanceSq = edg1.location.distanceSq(edg2.location);
                    edg1.setBotDistanceSq(botNames[y], distanceSq);
                    edg2.setBotDistanceSq(botNames[x], distanceSq);
                } else {
                    if (!edg1.alive) { edg2.removeDistanceSq(botNames[x]); }
                    if (!edg2.alive) { edg1.removeDistanceSq(botNames[y]); }
                }
            }
        }
    }
    
    protected Point2D.Double closestBotToEnemy(String enemyName) {
        Point2D.Double closestLocation = DiaUtils.nextLocation(_robot);;
        
        if (_enemiesAlive > 1) {
            EnemyDataGun edg = _enemies.get(enemyName);
            String closestEnemy = edg.closestBot();
            if (closestEnemy != null && 
                    edg.getBotDistanceSq(closestEnemy) <
                DiaUtils.square(_enemies.get(enemyName).distance)) {
                closestLocation = _enemies.get(closestEnemy).location;
            }
        }
        
        return closestLocation;
    }

    protected double cornerDistance(Point2D.Double p) {
        return Math.sqrt(
            DiaUtils.square(Math.min(p.x, _battleFieldWidth  - p.x)) +
            DiaUtils.square(Math.min(p.y, _battleFieldHeight - p.y)));
     }

    public void paintOn() {
        _painting = ENABLE_DEBUGGING_GRAPHICS;
    }
    
    public void paintOff() {
        _renderables.clear();
        _painting = false;
    }
    
    public boolean is1v1() {
        return (_enemiesAlive <= 1);
    }
    
    protected static final double[]  WEIGHTS_1V1 = 
        new double[]{4, 4, 4, 3, 4, 3, 3, 2, 2};
    protected static final double[]  WEIGHTS_FFA = 
        new double[]{3, 5, 2, 0, 4, 3, 0, 0, 2};

    // Produces a data point from one of our gun waves. This basically
    // defines how the gun aims.
    public double[] dataPointFromWave(DiaWave w, boolean aiming) {
        double[] point;
        point = new double[]{
            (Math.min(1000, w.targetDistance) / 1000),
//            (w.lateralVelocity() / 8),
            (Math.abs(w.targetVelocity) / 8),
            (w.targetRelativeHeading / Math.PI),
            (((w.targetAccel / (w.targetAccel < 0 ? Rules.DECELERATION : Rules.ACCELERATION)) + 1) / 2),
            (Math.min(1.0, w.targetWallDistance)),
            (Math.min(1.0, w.targetRevWallDistance)),
            (Math.min(1.25, w.targetVchangeTime / (w.targetDistance/w.bulletSpeed)) / 1.25),
//            (w.targetDl8t / 64),
//            (w.targetDl20t / 160),
//            (w.targetDl40t / 320),
            (aiming ? 0 : (Math.min(w.gunHeat, Rules.getGunHeat(w.bulletPower) - w.gunHeat) / (Rules.getGunHeat(w.bulletPower) / 2))),
            (((double)(_enemiesTotal - w.enemiesAlive)) / _enemiesTotal)
        };
        return point;
    }
}
