package voidious.gun;

import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

import voidious.utils.KdBucketTree;
import voidious.utils.DiaWave;

/*
 * General info about the enemy. We just keep one of these around for 
 * each enemy bot the gun is aware of.
 * 
 * The kd-tree and corresponding displacement vectors are used for actual
 * targeting.
 *  
 * Code is open source, released under the RoboWiki Public Code License:
 * http://robowiki.net/?RWPCL
 */

public class EnemyDataGun {
    public double distance;
    public double energy;
    public Point2D.Double location;
    public long lastScanTime;
    public ArrayList<Point2D.Double> pastLocations;
    public double damageTaken;
    public double damageGiven;
    public boolean alive;
    public double velocity;
    public double previousVelocity;
    public double heading;
    public double lastNonZeroVelocity;
    public double timeSinceDirectionChange;
    public double timeSinceVelocityChange;
    protected HashMap<String, Double> _botDistancesSq;
    public KdBucketTree scanTree;
    public HashMap<double[], Point2D.Double> displacementVectors;
    public DiaWave lastWaveFired;
    
    public EnemyDataGun(double dist, double en, Point2D.Double l, long time,
        double v, double h) {

        distance = dist;
        energy = en;
        location = l;
        lastScanTime = time;
        velocity = v;
        previousVelocity = 0;
        heading = h;
        damageTaken = 0;
        damageGiven = 0;
        alive = true;
        timeSinceDirectionChange = 0;
        timeSinceVelocityChange = 0;
        
        _botDistancesSq = new HashMap<String, Double>();
        scanTree = new KdBucketTree();
        displacementVectors = new HashMap<double[], Point2D.Double>();
        pastLocations = new ArrayList<Point2D.Double>();
        lastWaveFired = null;
    }
    
    public void saveDisplacementVector(double[] scanData, 
        Point2D.Double dispVector) {
        
        scanTree.insert(scanData);
        displacementVectors.put(scanData, dispVector);
    }
    
    public void setBotDistanceSq(String name, double distance) {
        _botDistancesSq.put(name, distance);
    }
    
    public double getBotDistanceSq(String name) {
        return _botDistancesSq.get(name);
    }
    
    public void clearDistancesSq() {
        _botDistancesSq.clear();
    }
    
    public void removeDistanceSq(String botName) {
        _botDistancesSq.remove(botName);
    }
    
    public String closestBot() {
        double min = Double.POSITIVE_INFINITY;
        String botName = null;
        
        Iterator<String> botNameIterator = _botDistancesSq.keySet().iterator();
        while (botNameIterator.hasNext()) {
            String name = botNameIterator.next();
            double dist = _botDistancesSq.get(name);
            if (dist < min) {
                min = dist;
                botName = name;
            }
        }

        return botName;
    }
    
    public Point2D.Double getPastLocation(int x) {
        return pastLocations.get(pastLocations.size() - 1 - x);
    }
}
