package voidious.move;

import java.awt.geom.Point2D;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.Vector;
import java.awt.Graphics2D;
import java.awt.Color;
import robocode.*;
import robocode.util.Utils;
import voidious.move.EnemyDataMove;
import voidious.utils.DiaUtils;
import voidious.utils.KdBucketTree;
import voidious.utils.RobotState;
import voidious.utils.DiaWave;
import voidious.utils.RoboGraphic;

/**
 * Code is open source, released under the RoboWiki Public Code License:
 * http://robowiki.net/?RWPCL
 */

//1v1 TODO: adjust for wall hits in enemy wave detection
//1v1 TODO: rammer detection

public class DiamondWhoosh {
    protected static final boolean ENABLE_DEBUGGING_GRAPHICS = true;
    protected static final double CURRENT_DESTINATION_BIAS = 0.8;
    protected static final int RECENT_LOCATIONS_TO_STORE = 5; 
    protected static final double DIRECTION_CHANGE_THRESHOLD = Math.PI / 2;
    protected static final long NUM_SLICES_BOT = 100;
//    protected static final long NUM_SLICES_PREVIOUS_DESTINATION = 25;
//    protected static final double MOVEMENT_STICK_PREVIOUS_DESTINATION = 50;
    protected static final double DEFAULT_ENERGY = 100;
    protected static final double LONGER_THAN_LONGEST_DISTANCE = 50000;
    protected static final double HALF_PI = Math.PI / 2;
    protected static final double BOT_HALF_WIDTH = 18;
    protected static final int INACTIVE_WAVE_OFFSET = -100;

    /////////////////////////////////////////////////////////////////////
    /// 1v1 stuff, delete what isn't needed later
    protected static final int WAVES_TO_SURF = 2;
    protected static final double DEFAULT_WALL_STICK = 160;
    protected static final double RAMMER_WALL_STICK = 130;
    protected static final int MAIN_CLUSTER_SIZE = 30;
    protected static final int RECENT_CLUSTER_SIZE = 20;
    protected static final int FLATTENER_CLUSTER_SIZE = 30;
    protected static final int RECENT_BULLET_HITS = 60; 
    protected static final double FEARFUL_DISTANCING_EXPONENT = 4;
    protected static final double NORMAL_DISTANCING_EXPONENT = 1.5;
    protected static final double SURF_ORIENTATION_VELOCITY_THRESHOLD = 0.1;
    protected static final double TYPICAL_DISTANCE = 465;
    protected static final double TYPICAL_ESCAPE_RANGE = 0.98;
    protected static final int CLOCKWISE_OPTION = 1;
    protected static final int STOP_OPTION = 0;
    protected static final int COUNTERCLOCKWISE_OPTION = -1;
    protected static final int NO_SURFABLE_WAVES = 0;
    protected static final int FIRST_WAVE = 0;
    protected static final int SECOND_WAVE = 1;
    protected static final DiaWave NO_WAVE_FOUND = null;
    protected static final boolean OBSERVE_WALL_HITS = false;
    protected double _desiredDistance;
    protected double _fearDistance;
    protected double _smoothAwayDistance;
    protected double _wallStick = DEFAULT_WALL_STICK;
    protected int _lastMovementChoice = 1;
    protected ArrayList<MovementChoice> _movementOptions;
    protected MovementChoice _optionCounterClockwise;
    protected MovementChoice _optionStop;
    protected MovementChoice _optionClockwise;
    protected DistanceController _subtleDistancer;
    protected DistanceController _aggressiveDistancer;
    protected DistanceController _currentDistancer;
    protected String _opponentName;
    protected ArrayList<double[][]> _cachedWaveNeighbors;
    protected DiaWave _lastWaveSurfed;
    protected double _lastVelocity;
    protected int _flattenerToggleTimer;
    /////////////////////////////////////////////////////////////////////
    
    protected AdvancedRobot _robot;
    protected HashMap<String, EnemyDataMove> _enemies; 
    protected Destination _currentDestination;
    protected double _currentHeading;
    protected double _previousHeading;
    protected long _timeSinceReverseDirection;
    protected long _timeSinceVelocityChange;
    protected int _minBotsCloser;
//    protected long _randomDirectionChangeTimer;

    protected Point2D.Double _myLocation;
    protected LinkedList<OldLocation> _recentLocations;
    protected double _myEnergy;
    protected long _currentTime;
    protected int _roundNum;
    protected int _enemiesAlive;
    protected int _enemiesTotal;
    protected double _lastNonZeroVelocity;
    protected LinkedList<DiaWave> _enemyWaves;
    protected LinkedList<DiaWave> _potentialWaves;
    protected LinkedList<Point2D.Double> _pastLocations;
    protected java.awt.geom.Rectangle2D.Double _fieldRect;
    protected double _battleFieldWidth;
    protected double _battleFieldHeight;
    
    protected Vector<RoboGraphic> _renderables;
    protected boolean _painting;
    protected Color _veryDarkGray = new Color(25, 25, 25);
    
    public DiamondWhoosh(AdvancedRobot robot) {
        _robot = robot;
        _enemies = new HashMap<String, EnemyDataMove>();
        _battleFieldWidth = _robot.getBattleFieldWidth();
        _battleFieldHeight = _robot.getBattleFieldHeight();
        _fieldRect = new java.awt.geom.Rectangle2D.Double(18, 18, 
            _battleFieldWidth - 36, _battleFieldHeight - 36);
        _recentLocations = new LinkedList<OldLocation>();
        _enemiesTotal = _robot.getOthers();
        _enemyWaves = new LinkedList<DiaWave>();
        _potentialWaves = new LinkedList<DiaWave>();
        _pastLocations = new LinkedList<Point2D.Double>();

        _currentDistancer = _subtleDistancer = new SubtleDistancing();
        _aggressiveDistancer = new AggressiveDistancing();
        _cachedWaveNeighbors = new ArrayList<double[][]>();
        _opponentName = "";
        initializeMovementOptions();
        
        _renderables = new Vector<RoboGraphic>();
        _painting = false;
    }

    public void initRound(AdvancedRobot robot) {
        _robot = robot;
        Iterator<EnemyDataMove> edmIterator = _enemies.values().iterator();
        while (edmIterator.hasNext()) {
            EnemyDataMove enemyData = edmIterator.next();
            enemyData.energy = DEFAULT_ENERGY;
            enemyData.distance = LONGER_THAN_LONGEST_DISTANCE;
            enemyData.alive = true;
            enemyData.clearDistancesSq();
            enemyData.lastTimeHit = Long.MIN_VALUE;
            enemyData.lastTimeClosest = Long.MIN_VALUE;
            enemyData.pastLocations.clear();
            enemyData.weighted1v1ShotsFiredThisRound = 0;
            enemyData.weighted1v1ShotsHitThisRound = 0;
        }
        _myLocation = new Point2D.Double(_robot.getX(), _robot.getY());
        _currentTime = 0;
        _roundNum = _robot.getRoundNum() + 1;
        _currentDestination = 
            new Destination(_myLocation, Double.POSITIVE_INFINITY, 0);
        _timeSinceReverseDirection = 0;
        _timeSinceVelocityChange = 0;
        _currentHeading = _previousHeading =
            _robot.getHeadingRadians();
        _enemiesAlive = _robot.getOthers();
        _lastNonZeroVelocity = 0;
        _minBotsCloser = 0;
        
        _recentLocations.clear();
        _enemyWaves.clear();
        _potentialWaves.clear();
        _pastLocations.clear();
        _cachedWaveNeighbors.clear();
        
        _renderables.clear();
    }
    
    public void execute() {
        _myLocation = new Point2D.Double(_robot.getX(), _robot.getY());
        _myEnergy = _robot.getEnergy();
        _currentTime = _robot.getTime();
        _previousHeading = _currentHeading;
        _enemiesAlive = _robot.getOthers();
        if (Math.abs(_robot.getVelocity()) > 
                SURF_ORIENTATION_VELOCITY_THRESHOLD) {
            _lastNonZeroVelocity = _robot.getVelocity();
        }
        _currentHeading = Utils.normalAbsoluteAngle(_robot.getHeadingRadians() + 
            (_lastNonZeroVelocity < 0 ? Math.PI : 0));
        Iterator<EnemyDataMove> edmIterator = _enemies.values().iterator();
        while (edmIterator.hasNext()) {
            EnemyDataMove enemyData = edmIterator.next();
            if (enemyData.lastScanTime < _currentTime) {
                enemyData.pastLocations.add(enemyData.location);
            }
        }
        updateBotDistances();
//        updateAvoidBeingTargeted();
//        updateStayPerpendicular();

        move();
        _myLocation = DiaUtils.nextLocation(_robot);
        _pastLocations.addFirst(_myLocation);
        _currentTime = _currentTime + 1;
    }
    
    protected void move() {
        if (is1v1()) {
            checkActiveEnemyWaves();
            move1v1();
        } else {
            updateTimeSinceTimers();
            moveFfa();
        }
    }
    
    public void move1v1() {
        evaluateDistancingControl();
//        evaluateWallStick();
        evaluateFlattener();
        
        surf();
    }
    
    public void moveFfa() {
        if (_currentTime % 7 == 0) {
            _recentLocations.addFirst(
                new OldLocation(DiaUtils.project(_myLocation, 
                    Math.random() * Math.PI * 2, 5 + Math.random() * 20), 
                    _currentTime));
            while (_recentLocations.size() > RECENT_LOCATIONS_TO_STORE) {
                _recentLocations.removeLast();
            }
        }
/*
        if (_timeSinceReverseDirection == 0) {
            _randomDirectionChangeTimer = Math.round(
                Math.random() * Math.sqrt(Math.random()) * 2 * 
                bulletTicksFromClosestBot()) - 1;
        }
*/
        if (_enemies.size() == 0) { return; }

        ArrayList<Destination> possibleDestinations = 
            new ArrayList<Destination>();
        
        possibleDestinations.addAll(generatePointsAroundBot());
//        possibleDestinations.addAll(generatePointsNearPreviousDestination());

        if (_myLocation.distance(_currentDestination.location) <=
            _myLocation.distance(_enemies.get(closestBot()).location)) {
            _currentDestination.goAngle = DiaUtils.absoluteBearing(_myLocation, 
                _currentDestination.location);
            _currentDestination.risk = CURRENT_DESTINATION_BIAS *
                evalDestinationRisk(_currentDestination.location, 
                    _currentDestination.goAngle);
            possibleDestinations.add(_currentDestination);
        }

        Destination nextDestination;
        do { 
            nextDestination = safestDestination(possibleDestinations);
            possibleDestinations.remove(nextDestination);
        } while (wouldHitWall(nextDestination));
            
        double goAngle = 
            DiaUtils.absoluteBearing(_myLocation, nextDestination.location);
        DiaUtils.setBackAsFront(_robot, goAngle);
        
        _currentDestination = nextDestination;

        if (_painting) {
            possibleDestinations.add(_currentDestination);
            drawRisks(possibleDestinations);
        }
    }

    public void onScannedRobot(ScannedRobotEvent e) {
        String botName = e.getName();
        _opponentName = botName;
        double absBearing = Utils.normalAbsoluteAngle(
                e.getBearingRadians() + _robot.getHeadingRadians());
        Point2D.Double enemyLocation = DiaUtils.project(_myLocation, 
            absBearing, e.getDistance());
        
        EnemyDataMove enemyData;
        double previousEnergy;
        if (_enemies.containsKey(botName)) {
            enemyData = _enemies.get(botName);
            previousEnergy = enemyData.energy;
            enemyData.energy = e.getEnergy();
            enemyData.distance = e.getDistance();
            enemyData.location = enemyLocation;
            enemyData.heading = e.getHeadingRadians();
            enemyData.absBearing = absBearing;
            enemyData.lastScanTime = _currentTime;
        } else {
            previousEnergy = e.getEnergy();
            enemyData = new EnemyDataMove(botName, e.getDistance(), 
                e.getEnergy(), enemyLocation, e.getHeadingRadians(), absBearing, 
                _currentTime);
            _enemies.put(botName, enemyData);
        }
        enemyData.pastLocations.add(enemyLocation);
        
        if (is1v1()) {
            updateTimeSinceTimers();
            if (_painting) {
                drawRawWaves();
            }

            double currentEnergy = e.getEnergy();
            boolean realShotFired = false;
            double bulletPower = previousEnergy - currentEnergy;
            if (bulletPower > 0.09 && bulletPower < 3.01) {
                realShotFired = true;
                enemyData.weighted1v1ShotsFired++;
                enemyData.weighted1v1ShotsFiredThisRound++;
            }
            fireEnemyWave(realShotFired, botName, bulletPower);
        }
    }
    
    public void onRobotDeath(RobotDeathEvent e) {
        String botName = e.getName();

        try {
            _enemies.get(botName).alive = false;
        } catch (java.lang.NullPointerException npe) {
            System.out.println(
                "WARNING (move): A bot died that I never knew existed!");
        }
    }
    
    public void onHitByBullet(HitByBulletEvent e) {
        String botName = e.getName();
        
        try {
            EnemyDataMove enemyData = _enemies.get(botName);
            enemyData.lastTimeHit = _robot.getTime();
            enemyData.damageTaken += 
               Rules.getBulletDamage(e.getBullet().getPower());
            enemyData.totalBulletPower += 
                (enemyData.lastBulletPower = e.getBullet().getPower());
            enemyData.totalTimesHit++;
            enemyData.energy += Rules.getBulletHitBonus(e.getBullet().getPower());
            DiaWave hitWave = 
                processBulletReturnFiringWave(e.getBullet(), e.getTime());
            if (hitWave != NO_WAVE_FOUND) {
                double thisHit = (hitWave.targetDistance
                    / TYPICAL_DISTANCE)
                    * (hitWave.escapeAngleRange() / TYPICAL_ESCAPE_RANGE);
                enemyData.weighted1v1ShotsHit += thisHit;
                enemyData.weighted1v1ShotsHitThisRound += thisHit;
            }
        } catch (java.lang.NullPointerException npe) {
            System.out.println(
                "WARNING (move): A bot shot me that I never knew existed!");
        }
        
        _cachedWaveNeighbors.clear();
    }
    
    public void onBulletHit(BulletHitEvent e) {
        String botName = e.getName();

        try {
            EnemyDataMove enemyData = _enemies.get(botName);
            enemyData.energy -= 
                Rules.getBulletDamage(e.getBullet().getPower());
            enemyData.damageGiven += 
                Rules.getBulletDamage(e.getBullet().getPower());
        } catch (java.lang.NullPointerException npe) {
            System.out.println(
                "WARNING (move): One of my bullets hit a bot that I never " +
                "knew existed!");
        }
        
    }

    public void onBulletHitBullet(BulletHitBulletEvent e) {
        String botName = e.getHitBullet().getName();

        try {
            EnemyDataMove enemyData = _enemies.get(botName);
            DiaWave hitWave = 
                processBulletReturnFiringWave(e.getHitBullet(), e.getTime());
            if (hitWave != NO_WAVE_FOUND) {
                enemyData.weighted1v1ShotsFired--; 
                enemyData.weighted1v1ShotsFiredThisRound--; 
            }
        } catch (java.lang.NullPointerException npe) {
            System.out.println(
                "WARNING (move): One of my bullets hit a bullet from a bot " +
                "that I never knew existed!");
        }

        _cachedWaveNeighbors.clear();
    }
    
    public void onPaint(Graphics2D g) {
        if (_painting) {
            Iterator<RoboGraphic> i = _renderables.iterator();
            while(i.hasNext()){
                RoboGraphic r = i.next();
                r.render(g);
            }
            _renderables.clear();
        }
    }
    
    public void onWin(WinEvent e) {
        roundOver();
    }
    
    public void onDeath(DeathEvent e) {
        roundOver();
    }
    
    protected void roundOver() {
        if (is1v1()) {
            duelOpponent().lastRoundNormalized1v1HitPercentage = 
                normalizedEnemyHitPercentageThisRound();
            System.out.println("Enemy normalized hit %: " + DiaUtils.round(normalizedEnemyHitPercentage(), 2));
            System.out.println("Flattener enabled: " + duelOpponent().flattenerEnabled);
        }
    }
    
    protected boolean wouldHitWall(Destination d) {
        long ticksAhead = 5;
        
        double heading = _robot.getHeadingRadians();
        double velocity = _robot.getVelocity();
        
        RobotState state = new RobotState(_myLocation, heading, velocity);
        boolean hitWall = false;
        
        for (int x = 0; x < ticksAhead && !hitWall; x++) {
            state = DiaUtils.nextLocation(state.location, state.velocity, 8.0, 
                state.heading, DiaUtils.absoluteBearing(state.location, 
                    d.location), 
                _currentTime + x, false, true, _battleFieldWidth, 
                _battleFieldHeight);
            if (!_fieldRect.contains(state.location)) {
                hitWall = true;
            }
        }
        
        return hitWall;
    }

    protected ArrayList<Destination> generatePointsAroundBot() {
        ArrayList<Destination> destinations = new ArrayList<Destination>();
        
        double movementStick = Math.min(100 + Math.random() * 100,
            distanceToClosestBot());
        
        double sliceSize = (2 * Math.PI) / NUM_SLICES_BOT;
        for (int x = 0; x < NUM_SLICES_BOT; x++) {
            double angle = x * sliceSize;
            Point2D.Double dest = DiaUtils.project(_myLocation, angle, 
                movementStick);
            if (_fieldRect.contains(dest)) {
                destinations.add(new Destination(dest, 
                evalDestinationRisk(dest, angle), angle));
            }
        }
        
        return destinations;
    }
/*
    protected ArrayList<Destination> generatePointsNearPreviousDestination() {
        ArrayList<Destination> destinations = new ArrayList<Destination>();
        
        double maxDistance = distanceToClosestBot();
        
        double sliceSize = (2 * Math.PI) / NUM_SLICES_PREVIOUS_DESTINATION;
        for (int x = 0; x < NUM_SLICES_PREVIOUS_DESTINATION; x++) {
            double angle = x * sliceSize;
            Point2D.Double dest = DiaUtils.project(_currentDestination.location, 
                angle, 10 + Math.random() * 40);
            if (_myLocation.distance(dest) <= maxDistance) {
                if (_fieldRect.contains(dest)) {
                    double angleFromBot = 
                        DiaUtils.absoluteBearing(_myLocation, dest);
                    destinations.add(new Destination(dest, 
                        evalDestinationRisk(dest, angleFromBot), angleFromBot));
                }
            }
        }
        
        return destinations;
    }
*/
    protected double evalDestinationRisk(Point2D.Double destination,
        double goAngle) {

        Iterator<EnemyDataMove> edmIterator = _enemies.values().iterator();
        double risk = 0;
        while (edmIterator.hasNext()) {
            EnemyDataMove enemyData = edmIterator.next();
            if (enemyData.alive) {
                double botRisk = 0;
                double distanceSq = destination.distanceSq(enemyData.location);
                botRisk = DiaUtils.limit(0.25, enemyData.energy / _myEnergy, 4)
                        * (1 + Math.abs(
                            Math.cos(enemyData.absBearing - goAngle)))
//                        * (20 + Math.sqrt(cornerDistance(destination)))
//                        * (_currentTime - enemyData.lastTimeHit < 100 ? 2 : 1)
//                        * (100 + enemyData.damageTaken) // could cache this
                        / (distanceSq
                            * (enemyData.botsCloser(distanceSq * .8) + 1)  
//                            * (100 + enemyData.damageGiven) // could cache this
//                            * (1 + enemyData.sumAbsBearingRisk(destination))
                            );
/*
                for (int x = 10; 
                     x < Math.min(101, enemyData.pastLocations.size()); 
                     x += 10) {
                    botRisk *= 1 + (250 / enemyData.getPastLocation(x)
                                .distanceSq(destination));
                }
*/
                risk += botRisk;
            }
        }

        Iterator<OldLocation> rlIterator = _recentLocations.iterator();
        while (rlIterator.hasNext()) {
            OldLocation ol = rlIterator.next();
            risk *= 1 + (500.0 / ol.location.distanceSq(destination));
        }

/*
        boolean wouldChangeDirection = 
            Math.abs(Utils.normalRelativeAngle(goAngle - 
                _currentHeading)) > DIRECTION_CHANGE_THRESHOLD;
        if (_enemiesAlive <= 6 &&
            ((wouldChangeDirection && 
            _timeSinceReverseDirection < _randomDirectionChangeTimer) ||
            (!wouldChangeDirection &&
            _timeSinceReverseDirection >= _randomDirectionChangeTimer))) {
            risk *= 2;
        }
*/
        return risk;
    }
    
    protected Destination safestDestination(
        ArrayList<Destination> possibleDestinations) {
        
        double lowestRisk = Double.POSITIVE_INFINITY;
        Destination safest = null;
        
        Iterator<Destination> destIterator = possibleDestinations.iterator();
        while (destIterator.hasNext()) {
            Destination dest = destIterator.next();
            if (dest.risk < lowestRisk) {
                lowestRisk = dest.risk;
                safest = dest;
            }
        }
        
        if (safest == null) {
            System.out.println("WARNING: No safe destinations found, there " +
                "must be a bug in the risk evaluation.");
            safest = _currentDestination;
        }

        return safest;
    }
/*    
    protected double cornerDistance(Point2D.Double p) {
        return Math.sqrt(
            DiaUtils.square(Math.min(p.x, _battleFieldWidth  - p.x)) +
            DiaUtils.square(Math.min(p.y, _battleFieldHeight - p.y)));
     }
    
    protected double nearestWallDistance(Point2D.Double p) {
        return Math.min(p.y, Math.min(_battleFieldHeight - p.y, 
            Math.min(p.x, _battleFieldWidth  - p.x)));
    }
*/
    protected double bulletTicksFromClosestBot() {
        if (_enemies.isEmpty()) { return 10; }
        
        return bulletTicksFromBot(closestBot());
    }
    
    protected double bulletTicksFromBot(String botName) {
        EnemyDataMove ms = _enemies.get(botName);
        double bulletTicks = 
            Math.ceil((ms.distance - BOT_HALF_WIDTH)
                / Rules.getBulletSpeed(ms.avgBulletPower()));
        
        return bulletTicks;
    }
    
    protected String closestBot() {
        double closestDistance = Double.POSITIVE_INFINITY;
        String closestBot = null;
        
        Iterator<String> enemyDataKeys = _enemies.keySet().iterator();
        while (enemyDataKeys.hasNext()) {
            String botName = enemyDataKeys.next();
            EnemyDataMove enemyData = _enemies.get(botName);
            if (enemyData.alive && enemyData.distance < closestDistance) {
                closestDistance = enemyData.distance;
                closestBot = botName;
            }
        }
        
        return closestBot;
    }
    
    protected double distanceToClosestBot() {
        double closestDistance = Double.POSITIVE_INFINITY;
        
        Iterator<String> enemyDataKeys = _enemies.keySet().iterator();
        while (enemyDataKeys.hasNext()) {
            String botName = enemyDataKeys.next();
            EnemyDataMove enemyData = _enemies.get(botName);
            if (enemyData.alive && enemyData.distance < closestDistance) {
                closestDistance = enemyData.distance;
            }
        }
        
        return closestDistance;
    }

    protected void updateBotDistances() {
        if (_enemies.size() <= 1) { return; }

        String[] botNames = new String[_enemies.size()];
        _enemies.keySet().toArray(botNames);
        
        for (int x = 0; x < botNames.length; x++) {
            EnemyDataMove edm1 = _enemies.get(botNames[x]);
            for (int y = x + 1; y < botNames.length; y++) {
                EnemyDataMove edm2 = _enemies.get(botNames[y]);
                if (edm1.alive && edm2.alive) {
                    double distanceSq = edm1.location.distanceSq(edm2.location);
                    edm1.setBotDistanceSq(botNames[y], distanceSq);
                    edm2.setBotDistanceSq(botNames[x], distanceSq);
                } else {
                    if (!edm1.alive) { edm2.removeDistanceSq(botNames[x]); }
                    if (!edm2.alive) { edm1.removeDistanceSq(botNames[y]); }
                }
            }
            edm1.distance = _myLocation.distance(edm1.location);
        }
    }
/*
    protected void updateAvoidBeingTargeted() {
        Iterator<EnemyDataMove> edmIterator = _enemies.values().iterator();
        _minBotsCloser = _enemiesTotal;
        while (edmIterator.hasNext()) {
            EnemyDataMove enemyData = edmIterator.next();
            double distanceSq = 
                _myLocation.distanceSq(enemyData.location);
            if (enemyData.damageTaken * 1.5 >= enemyData.damageGiven) {
                enemyData.avoidBeingTargeted = true;
            } else {
                enemyData.avoidBeingTargeted = false;
            }
            int botsCloser = enemyData.botsCloser(distanceSq * .8); 
            if (botsCloser < _minBotsCloser) {
                _minBotsCloser = botsCloser;
            }
        }
    }

    protected void updateStayPerpendicular() {
        Iterator<EnemyDataMove> edmIterator = _enemies.values().iterator();
        while (edmIterator.hasNext()) {
            EnemyDataMove enemyData = edmIterator.next();
            double distanceSq = 
                _myLocation.distanceSq(enemyData.location);
            int botsCloser = enemyData.botsCloser(distanceSq); 
            if (botsCloser == 0) {
                enemyData.lastTimeClosest = _currentTime;
            }
            if (_currentTime - enemyData.lastTimeHit < 75 ||
                _currentTime - enemyData.lastTimeClosest < 100 ||
                botsCloser <= 1) {
                enemyData.stayPerpendicular = true;
            } else {
                enemyData.stayPerpendicular = false;
            }
        }
    }
*/    
    public void surf() {
        if (_opponentName.equals("")) {
            return;
        }
        
        RobotState currentState = new RobotState(_myLocation,
            _robot.getHeadingRadians(), _robot.getVelocity(), _robot.getTime());
        boolean goingClockwise = (_lastMovementChoice == CLOCKWISE_OPTION);
            Collections.sort(_movementOptions);
        
        Collections.sort(_movementOptions);

        double bestSurfDanger = Double.POSITIVE_INFINITY;
        for (int x = 0; x < _movementOptions.size(); x++) {
            MovementChoice mc = _movementOptions.get(x);
            double thisDanger = checkDanger(currentState, 
                mc.getMovementOption(), goingClockwise, FIRST_WAVE, 
                WAVES_TO_SURF, bestSurfDanger);
            mc.lastDanger = thisDanger;
            
            bestSurfDanger = Math.min(bestSurfDanger, thisDanger);
        }
        
        double orbitCounterClockwiseDanger = _optionCounterClockwise.lastDanger;
        double stopDanger = _optionStop.lastDanger;
        double orbitClockwiseDanger = _optionClockwise.lastDanger;
        
        int goOrientation = _lastMovementChoice;

        DiaWave orbitWave = findSurfableWave(FIRST_WAVE);
        if (orbitWave != _lastWaveSurfed) {
            _cachedWaveNeighbors.clear();
            _lastWaveSurfed = orbitWave;
        }
        
        double orbitAbsBearing, distanceToClosestWaveSource;
        try {
            distanceToClosestWaveSource = _myLocation
                .distance(orbitWave.sourceLocation);
            orbitAbsBearing = DiaUtils.absoluteBearing(orbitWave.sourceLocation,
                _myLocation);
        } catch (NullPointerException noSurfableWaves) {
            distanceToClosestWaveSource = _enemies.get(_opponentName).distance;
            orbitAbsBearing = _enemies.get(_opponentName).absBearing + Math.PI;
        }
        
        double goAngle, attackAngle;
        if (stopDanger == NO_SURFABLE_WAVES) {
            attackAngle = -1.047;
            _robot.setMaxVelocity(8);
            
            double goAngleCcw = orbitAbsBearing +
                (COUNTERCLOCKWISE_OPTION * ((Math.PI / 2) + attackAngle));
            goAngleCcw = wallSmoothing(_myLocation, goAngleCcw, 
                COUNTERCLOCKWISE_OPTION, distanceToClosestWaveSource);

            double goAngleCw = orbitAbsBearing +
                (CLOCKWISE_OPTION * ((Math.PI / 2) + attackAngle));
            goAngleCw = wallSmoothing(_myLocation, goAngleCw, 
                CLOCKWISE_OPTION, distanceToClosestWaveSource);
            
            if (Math.abs(Utils.normalRelativeAngle(goAngleCw - orbitAbsBearing))
                < Math.abs(Utils.normalRelativeAngle(goAngleCcw - orbitAbsBearing))) {
                goOrientation = CLOCKWISE_OPTION;
                goAngle = goAngleCw;
            } else {
                goOrientation = COUNTERCLOCKWISE_OPTION;
                goAngle = goAngleCcw;
            }
        } else {
            _robot.setMaxVelocity(8);
            attackAngle = 
                _currentDistancer.attackAngle(distanceToClosestWaveSource, 
                    _desiredDistance);

            if (stopDanger <= orbitCounterClockwiseDanger &&
                stopDanger <= orbitClockwiseDanger) {
                
                _robot.setMaxVelocity(0);
            } else {            
                if (orbitClockwiseDanger < orbitCounterClockwiseDanger) {
                    goOrientation = CLOCKWISE_OPTION;
                } else {
                    goOrientation = COUNTERCLOCKWISE_OPTION;
                }
            }
            
            goAngle = orbitAbsBearing +
                (goOrientation * ((Math.PI / 2) + attackAngle));
            goAngle = wallSmoothing(_myLocation, goAngle, goOrientation,
                distanceToClosestWaveSource);

            if (_painting) {
                for (int x = 0; x < WAVES_TO_SURF; x++) {
                    drawWaveDangers(x);
                }
            }
        }
        
        
        DiaUtils.setBackAsFront(_robot, goAngle);
        
        _lastMovementChoice = goOrientation;
        
    }

    public double checkDanger(RobotState startState, int movementOption, 
        boolean previouslyMovingClockwise, int surfableWaveIndex, 
        int recursionLevels, double cutoffDanger) {
            
        if (surfableWaveIndex >= recursionLevels) {
            return 0;
        }

        boolean predictClockwiseOrNot;
        if (movementOption == CLOCKWISE_OPTION) {
            predictClockwiseOrNot = true;
        } else if (movementOption == COUNTERCLOCKWISE_OPTION) {
            predictClockwiseOrNot = false;
        } else {
            predictClockwiseOrNot = previouslyMovingClockwise;
        }

        DiaWave surfWave = findSurfableWave(surfableWaveIndex);
        
        if (surfWave == null) {
/*
            if (surfableWaveIndex == FIRST_WAVE) {
                double nonSurfableWaveDistance = 150;
                surfWave = findNonSurfableWave(nonSurfableWaveDistance);
            }
*/
            if (surfWave == null) { return NO_SURFABLE_WAVES; }
        }
      
//        Color drawColor = Color.white;
//        if (surfableWaveIndex != 0 || movementOption == STOP_OPTION) {
//            drawColor = Color.blue;         
//        }
      
        double waveHitInterceptOffset = 
            surfWave.bulletSpeed + BOT_HALF_WIDTH;
        double wavePassedInterceptOffset = surfWave.bulletSpeed;
        RobotState predictedState = startState;
        RobotState dangerState = startState;
        
        boolean wavePassed = false;
        boolean waveHit = false;
        
        double maxVelocity = (movementOption == STOP_OPTION) ? 0 : 8;
        
        do {    
            double orbitAbsBearing = DiaUtils.absoluteBearing(
                surfWave.sourceLocation, predictedState.location);
            double orbitDistance = 
                surfWave.sourceLocation.distance(predictedState.location);
            double attackAngle = _currentDistancer.attackAngle(orbitDistance, 
                _desiredDistance);
            boolean clockwiseSmoothing = predictClockwiseOrNot;

            if (orbitDistance < _smoothAwayDistance) {
                clockwiseSmoothing = !clockwiseSmoothing;
            }
            
            predictedState = 
                DiaUtils.nextPerpendicularWallSmoothedLocation(
                    predictedState.location, orbitAbsBearing, 
                    predictedState.velocity, maxVelocity, 
                    predictedState.heading, attackAngle, clockwiseSmoothing, 
                    predictedState.time, _fieldRect, _battleFieldWidth, 
                    _battleFieldHeight, _wallStick, 
                    OBSERVE_WALL_HITS);
            
            if (!waveHit &&
                surfWave.wavePassed1v1(predictedState.location, 
                    predictedState.time, waveHitInterceptOffset)) {
                
                dangerState = predictedState;
                waveHit = true;
            }
            
            if (!wavePassed &&
                surfWave.wavePassed1v1(predictedState.location, 
                    predictedState.time, wavePassedInterceptOffset)) {

                wavePassed = true;              
            }
        } while (!wavePassed);
        
//        _renderables.add(RoboGraphic.drawPoint(
//            predictedState.location, drawColor));

        double danger = getDangerScore(surfWave, dangerState.location,
            surfableWaveIndex);
        danger *= Rules.getBulletDamage(surfWave.bulletPower); 

        double currentDistanceToWaveSource = 
            _myLocation.distance(surfWave.sourceLocation);
        double currentDistanceToWave = 
            currentDistanceToWaveSource - 
                surfWave.distanceTraveled(_robot.getTime());
        double timeToImpact = currentDistanceToWave 
            / Rules.getBulletSpeed(surfWave.bulletPower);
        
//        if (duelOpponent().flattenerEnabled) {
//            danger /= DiaUtils.square(timeToImpact);          
//        } else {
            danger /= timeToImpact;
//        }

        double firstWaveMultiplier = 1;
        if (surfableWaveIndex == FIRST_WAVE) {
            double predictedDistanceToWaveSource = 
                surfWave.sourceLocation.distance(predictedState.location);
            double predictedDistanceToEnemy =
                _enemies.get(_opponentName).location
                    .distance(predictedState.location);
            
            double shorterDistance = Math.min(predictedDistanceToWaveSource,
                predictedDistanceToEnemy);
            double distancingDangerBase = 
                Math.max(currentDistanceToWaveSource / shorterDistance, .99);
            double distancingDangerExponent = shorterDistance > _fearDistance ?
                NORMAL_DISTANCING_EXPONENT : FEARFUL_DISTANCING_EXPONENT;

            firstWaveMultiplier = Math.pow(distancingDangerBase, 
                distancingDangerExponent);          
        }

        if (danger * firstWaveMultiplier > cutoffDanger) {
            return danger * firstWaveMultiplier;
        }
        
        double nextCounterClockwiseDanger = checkDanger(predictedState,
            COUNTERCLOCKWISE_OPTION, predictClockwiseOrNot, 
            surfableWaveIndex + 1, recursionLevels, cutoffDanger);
        double nextStopDanger = checkDanger(predictedState, 
            STOP_OPTION, predictClockwiseOrNot, surfableWaveIndex + 1,
            recursionLevels, cutoffDanger);
        double nextClockwiseDanger = checkDanger(predictedState,
            CLOCKWISE_OPTION, predictClockwiseOrNot, 
            surfableWaveIndex + 1, recursionLevels, cutoffDanger);
        
        danger += Math.min(nextCounterClockwiseDanger,
            Math.min(nextStopDanger, nextClockwiseDanger));

        danger *= firstWaveMultiplier;      
        
        return danger;
    }
     
    public double getDangerScore(DiaWave w, Point2D.Double predictedLocation,
        int surfableWaveIndex) {
        EnemyDataMove enemyData = _enemies.get(w.botName);
        double checkFactor = w.guessFactor(predictedLocation);

        if (enemyData.guessFactors.size() == 0) {
            // HoT avoidance
            return 1 - DiaUtils.square(checkFactor);
        }
        
        double[] weights = normalizedEnemyHitPercentage() > 4 ?
            WEIGHTS_NORMAL : WEIGHTS_SIMPLE;

        KdBucketTree scanTree = enemyData.scanTree;
        double[] wavePoint = dataPointFromWave(w);
        double[][] nearestNeighbors;
        
        if (_cachedWaveNeighbors.size() > surfableWaveIndex) {
            nearestNeighbors = _cachedWaveNeighbors.get(surfableWaveIndex);
        } else {
            nearestNeighbors = KdBucketTree.nearestNeighbors(
                scanTree, wavePoint, 
                Math.min((int)Math.ceil(enemyData.guessFactors.size() / 5.0), 
                    MAIN_CLUSTER_SIZE), 
                weights);
            
            if (!enemyData.recentDataPoints.isEmpty() && 
                    normalizedEnemyHitPercentage() > 7.5) {
                double[][] recentNeighbors = KdBucketTree.nearestNeighbors(
                    enemyData.recentScanTree, wavePoint, 
                    Math.min(enemyData.recentDataPoints.size(), 
                        RECENT_CLUSTER_SIZE),
                    weights);
                
                double[][] combinedNeighbors = 
                    new double[nearestNeighbors.length + recentNeighbors.length]
                              [wavePoint.length];
                for (int x = 0; x < nearestNeighbors.length; x++) {
                    combinedNeighbors[x] = nearestNeighbors[x];
                }
                
                int recentOffset = nearestNeighbors.length;
                for (int x = 0; x < recentNeighbors.length; x++) {
                    combinedNeighbors[x + recentOffset] = recentNeighbors[x];
                }
                
                nearestNeighbors = combinedNeighbors;
            }

            if (duelOpponent().flattenerEnabled) { 
                double[][] flatNeighbors = KdBucketTree.nearestNeighbors(
                    enemyData.flattenerTree, wavePoint, 
                    Math.min(enemyData.flatFactors.size(), 
                        FLATTENER_CLUSTER_SIZE), 
                    weights);
                
                double[][] combinedNeighbors = 
                    new double[nearestNeighbors.length + flatNeighbors.length]
                              [wavePoint.length];
                for (int x = 0; x < nearestNeighbors.length; x++) {
                    combinedNeighbors[x] = nearestNeighbors[x];
                }
                
                int flatOffset = nearestNeighbors.length;
                for (int x = 0; x < flatNeighbors.length; x++) {
                    combinedNeighbors[x + flatOffset] = flatNeighbors[x];
                }
                
                nearestNeighbors = combinedNeighbors;
            }

            _cachedWaveNeighbors.add(surfableWaveIndex, nearestNeighbors);
        }
        
        int numScans = nearestNeighbors.length;

        if (numScans == 0) {
            // HoT avoidance
            return 1 - DiaUtils.square(checkFactor);
        }

        double[] guessFactors = new double[numScans];
        for (int x = 0; x < numScans; x++) {
            if (duelOpponent().flattenerEnabled && 
                enemyData.flatFactors.containsKey(nearestNeighbors[x])) {
                guessFactors[x] = 
                    enemyData.flatFactors.get(nearestNeighbors[x]);
            } else {
                guessFactors[x] = enemyData.guessFactors.get(nearestNeighbors[x]);
            }
        }
        
        double density = 0;
        double bandwidth = DiaUtils.botWidthAimAngle(
            predictedLocation.distance(w.sourceLocation))
                / w.preciseEscapeAngle(checkFactor >= 0);
        for (int x = 0; x < numScans; x++) {
            double xGuessFactor = guessFactors[x];
            
            double ux = (xGuessFactor - checkFactor) / bandwidth;

            // Gaussian
            density += Math.exp(-0.5 * ux * ux)
                / KdBucketTree.distance(nearestNeighbors[x], wavePoint);
        }

        return density;
    }
    
    public void initializeMovementOptions() {
        _movementOptions = new ArrayList<MovementChoice>();
        
        _movementOptions.add(_optionCounterClockwise = 
            new MovementCounterClockwise());
        _movementOptions.add(_optionStop = new MovementStop());
        _movementOptions.add(_optionClockwise = new MovementClockwise());
        
    }
    
    public void fireEnemyWave(boolean detectedEnemyWave, String botName,
        double bulletPower) {

        double currentVelocity = _robot.getVelocity();
        EnemyDataMove enemyData = _enemies.get(botName);
        int velocitySign = DiaUtils.nonZeroSign(
            Math.abs(currentVelocity) > SURF_ORIENTATION_VELOCITY_THRESHOLD ? 
                currentVelocity : _lastNonZeroVelocity);
        DiaWave futureWave = new DiaWave(enemyData.location, _myLocation,
            enemyData.location, _robot.getTime() + 1, 3.0, botName,
            _robot.getHeadingRadians(), currentVelocity, 0, velocitySign, 
            enemyData.distance, _timeSinceReverseDirection, 
            _timeSinceVelocityChange, 0, 0, 
            _myLocation.distance(
                _pastLocations.get(Math.min(8, _pastLocations.size()-1))),
            _myLocation.distance(
                _pastLocations.get(Math.min(20, _pastLocations.size()-1))),
            _myLocation.distance(
                _pastLocations.get(Math.min(40, _pastLocations.size()-1))),
            1, 0, _fieldRect, _battleFieldWidth, _battleFieldHeight);
        _potentialWaves.addFirst(futureWave);
        
        if (detectedEnemyWave && _potentialWaves.size() >= 3) {
            DiaWave realWave = _potentialWaves.get(2);
            Point2D.Double aimedFromLocation =
                enemyData.getPastLocation(2);
            realWave.sourceLocation = enemyData.getPastLocation(1);
            realWave.targetLocation = _pastLocations.get(2);
            realWave.absBearing = 
                DiaUtils.absoluteBearing(aimedFromLocation, 
                    realWave.targetLocation);
            realWave.setBulletPower(bulletPower);
            realWave.targetWallDistance =
                Math.min(1.5, DiaUtils.orbitalWallDistance(
                    aimedFromLocation, realWave.targetLocation, 
                    bulletPower, realWave.orbitDirection, _fieldRect));
            realWave.targetRevWallDistance =
                Math.min(1.5, DiaUtils.orbitalWallDistance(
                    aimedFromLocation, realWave.targetLocation, 
                    bulletPower, -realWave.orbitDirection, _fieldRect));
            _enemyWaves.addLast(realWave);
        }
    }
    
    public void checkActiveEnemyWaves() {
        long currentTime = _robot.getTime();

        Iterator<DiaWave> wavesIterator = _enemyWaves.iterator();
        while (wavesIterator.hasNext()) {
            DiaWave w = wavesIterator.next();
            if (!w.processedWaveBreak &&
                    w.wavePassed1v1(_myLocation, currentTime, 0)) {
                EnemyDataMove enemyData = _enemies.get(w.botName);
                double[] wavePoint = dataPointFromWave(w);
                enemyData.flattenerTree.insert(wavePoint);
                enemyData.flatFactors.put(wavePoint, 
                    w.guessFactor(_myLocation));
                w.processedWaveBreak = true;
            }
            if (w.wavePassed1v1(_myLocation, currentTime, 
                    INACTIVE_WAVE_OFFSET)) {
                wavesIterator.remove();
            }
        }
        
    }
    
    public DiaWave processBulletReturnFiringWave(Bullet bullet, 
        long currentTime) {
            
        Point2D.Double bulletLocation = 
            new Point2D.Double(bullet.getX(), bullet.getY());
        String botName = bullet.getName();
        
        int tightMatchDistanceThreshold = 50;
        DiaWave hitWave = DiaUtils.findClosestWave(_enemyWaves,
            bulletLocation, _robot.getTime(), DiaWave.ANY_WAVE,
            tightMatchDistanceThreshold);
        
        if (hitWave == null) {
            return NO_WAVE_FOUND; 
        }
        
        if (DiaUtils.round(bullet.getPower(), 1) != 
                DiaUtils.round(hitWave.bulletPower, 1) ||
            !botName.equals(hitWave.botName)) {
            return NO_WAVE_FOUND;
        }
        
        double hitGuessFactor = hitWave.guessFactor(bulletLocation);
        EnemyDataMove enemyData = _enemies.get(botName);
        double[] dataPoint = dataPointFromWave(hitWave);
        enemyData.scanTree.insert(dataPoint);
        enemyData.recentScanTree.insert(dataPoint);
        enemyData.guessFactors.put(dataPoint, hitGuessFactor);
        enemyData.recentDataPoints.addLast(dataPoint);
        while (enemyData.recentDataPoints.size() > RECENT_BULLET_HITS) {
            double[] oldPoint = enemyData.recentDataPoints.getFirst();
            enemyData.recentScanTree.remove(oldPoint);
            enemyData.recentDataPoints.removeFirst();
        }
        
        if (_painting) {
            _renderables.add(RoboGraphic.drawLine(
                hitWave.sourceLocation,
                DiaUtils.project(hitWave.sourceLocation, hitWave.absBearing, 
                    hitWave.distanceTraveled(currentTime)), 
                Color.yellow));
            if (Math.abs(hitGuessFactor) > 0.01) {
                _renderables.add(RoboGraphic.drawLine(
                    hitWave.sourceLocation, bulletLocation, 
                    (hitGuessFactor > 0 ? Color.blue : Color.red)));
            }
            
        }

        hitWave.processedBulletHit = true;
        
        return hitWave;     
    }
    
    public DiaWave findSurfableWave(int waveIndex) {
        int searchWaveIndex = 0;
        long currentTime = _robot.getTime();
        
        Iterator<DiaWave> wavesIterator = _enemyWaves.iterator();
        while (wavesIterator.hasNext()) {
            DiaWave w = wavesIterator.next();
            double distanceToWaveSource = 
                _myLocation.distance(w.sourceLocation);
            double distanceToWave = distanceToWaveSource -
                w.distanceTraveled(currentTime);
            
            if (!w.processedBulletHit && distanceToWave > w.bulletSpeed) {
                if (searchWaveIndex == waveIndex) {
                    return w;
                } else {
                    searchWaveIndex++;
                }
            }
        }
        
        return NO_WAVE_FOUND;
    }
    
    public void evaluateDistancingControl() {
        double normalizedHitPercentageThreshold;

        if (_robot.getRoundNum() < 2) {
            _desiredDistance = 525;
            _fearDistance = 175;
            _smoothAwayDistance = 75;
        } else if (normalizedEnemyHitPercentage() 
                > (normalizedHitPercentageThreshold = 7.5)) {
            _desiredDistance = 
                525 + (Math.max((duelOpponent().avgBulletPower() - 2), 0) * 165);
            _fearDistance = 250;
            _smoothAwayDistance = 75;
        } else if (normalizedEnemyHitPercentage() 
                > (normalizedHitPercentageThreshold = 3)) {
            _desiredDistance = 
                450 + (Math.max((duelOpponent().avgBulletPower() - 2), 0) * 150);
            _fearDistance = 175;
            _smoothAwayDistance = 75;
        } else {
            _desiredDistance = 450;
            _fearDistance = 0;
            _smoothAwayDistance = 0;
        }
        
        if (_robot.getRoundNum() < 2 || 
            normalizedEnemyHitPercentage() > 7.5) {
            _currentDistancer = _aggressiveDistancer;
        } else {
            _currentDistancer = _subtleDistancer;
        }
        
        _desiredDistance = 550;
        _fearDistance = 250;
        _smoothAwayDistance = 75;
    }
    
    public void evaluateFlattener() {
        EnemyDataMove enemyData = duelOpponent();
        if (enemyData == null) { return; }

        double normalizedHitPercentageThreshold =
            calculateFlattenerThreshold();
        
        double thresholdMargin = (enemyData.flattenerEnabled ? -1.0 : 0.0);
        if (normalizedEnemyHitPercentage() > 
            normalizedHitPercentageThreshold + thresholdMargin) {
            setFlattener(true);
        } else {
            setFlattener(false);
        }
    }

    public double calculateFlattenerThreshold() {
        double normalizedHitPercentageThreshold;
        
        normalizedHitPercentageThreshold = 200;;
        
        EnemyDataMove enemyData = duelOpponent();
        if (_robot.getRoundNum() >= 20 && 
                enemyData.weighted1v1ShotsFired > 500) {
            normalizedHitPercentageThreshold = 10.0;
        } else if (_robot.getRoundNum() >= 10 &&
                enemyData.weighted1v1ShotsFired > 300) {
            normalizedHitPercentageThreshold = 10.7;
        } else if (_robot.getRoundNum() >= 5 &&
                enemyData.weighted1v1ShotsFired > 150) {
            if (normalizedEnemyHitPercentageLastRound() >= 12) {
                normalizedHitPercentageThreshold = 12;
            } else {
                normalizedHitPercentageThreshold = 15;         
            }
        }
        
        return normalizedHitPercentageThreshold;
    }
    
    public double normalizedEnemyHitPercentage() { 
        EnemyDataMove enemyData = duelOpponent();

        return ((enemyData == null || 
                    enemyData.weighted1v1ShotsFired == 0) ? 0 : 
            (((double)enemyData.weighted1v1ShotsHit / 
                enemyData.weighted1v1ShotsFired) * 100.0)); 
    }

    public double normalizedEnemyHitPercentageThisRound() { 
        EnemyDataMove enemyData = duelOpponent();

        return ((enemyData == null ||
                    enemyData.weighted1v1ShotsFiredThisRound == 0) ? 0 : 
            (((double)enemyData.weighted1v1ShotsHitThisRound / 
                enemyData.weighted1v1ShotsFiredThisRound) * 100.0)); 
    }

    public double normalizedEnemyHitPercentageLastRound() {
        EnemyDataMove enemyData = duelOpponent();

        return enemyData == null ? 0 : 
            enemyData.lastRoundNormalized1v1HitPercentage;
    }
    
    public void setFlattener(boolean enableFlattener) {
        boolean oldFlattenerSetting = duelOpponent().flattenerEnabled;
        
        int preventFlattenerThrashing = 100;
        if (_flattenerToggleTimer++ > preventFlattenerThrashing && 
            oldFlattenerSetting != enableFlattener) {
            _flattenerToggleTimer = 0;
            
            if (enableFlattener) { enableFlattener(); } 
            else { disableFlattener(); }
            _cachedWaveNeighbors.clear();
        }
    }
    
    public void enableFlattener() {
        duelOpponent().flattenerEnabled = true;
        
        System.out.println("Curve Flattening enabled.");
    }

    public void disableFlattener() {
        duelOpponent().flattenerEnabled = false;

        System.out.println("Curve Flattening disabled.");
    }
    
    public double wallSmoothing(Point2D.Double startLocation, 
        double goAngleRadians, int orientation, double currentDistance) {
            
        if (currentDistance < _smoothAwayDistance) {
            orientation *= -1;
        }
        
        double smoothedAngle = DiaUtils.wallSmoothing(
            _fieldRect, _battleFieldWidth, _battleFieldHeight, startLocation, 
            goAngleRadians, orientation, _wallStick);

        return smoothedAngle;
    }

    // In FFA, fine to do this from execute().
    // In 1v1, want to do it before firing waves, so we do
    // it from onScannedRobot.
    public void updateTimeSinceTimers() {
        if (Math.abs(Utils.normalRelativeAngle(_currentHeading - 
                _previousHeading)) > DIRECTION_CHANGE_THRESHOLD) {
            _timeSinceReverseDirection = 0;
        } else {
            _timeSinceReverseDirection++;
        }
        
        double newVelocity = _robot.getVelocity();
        if (Math.abs(newVelocity - _lastVelocity) > 0.5) {
            _timeSinceVelocityChange = 0;
        } else {
            _timeSinceVelocityChange++;
        }
        _lastVelocity = newVelocity;
    }

    public void drawRisks(ArrayList<Destination> destinations) {
        double lowestRisk = Double.POSITIVE_INFINITY;
        double highestRisk = Double.NEGATIVE_INFINITY;
        
        Iterator<Destination> destIterator = destinations.iterator();
        double[] risks = new double[destinations.size()];
        int x = 0;
        while (destIterator.hasNext()) {
            Destination d = destIterator.next();
            risks[x++] = d.risk; 
            if (d.risk < lowestRisk) {
                lowestRisk = d.risk;
            }
            
            if (d.risk > highestRisk) {
                highestRisk = d.risk;
            }
        }
        
        
        double avg = DiaUtils.average(risks);
        double stDev = DiaUtils.standardDeviation(risks);

        destIterator = destinations.iterator();
        
        while (destIterator.hasNext()) {
            Destination d = destIterator.next();
            
            _renderables.add(RoboGraphic.drawCircleFilled(d.location, 
                riskColor(d.risk - lowestRisk, avg - lowestRisk, 
                    stDev, true, 1), 2));
        }
    }
    
    public void drawRawWaves() {
        long currentTime = _robot.getTime();
        Iterator<DiaWave> wavesIterator = _enemyWaves.iterator();
        while (wavesIterator.hasNext()) {
            DiaWave w = wavesIterator.next();
            _renderables.add(RoboGraphic.drawCircle(w.sourceLocation, 
                (currentTime - w.fireTime) * w.bulletSpeed, _veryDarkGray));
            _renderables.add(RoboGraphic.drawLine(w.sourceLocation, 
                DiaUtils.project(w.sourceLocation, w.absBearing, 
                    w.distanceTraveled(currentTime)), _veryDarkGray));
            _renderables.add(RoboGraphic.drawCircle(
                DiaUtils.project(w.sourceLocation, w.absBearing, 
                    w.distanceTraveled(currentTime)),
                5, _veryDarkGray));
        }
    }
    
    // TODO: why convert to location only to convert back to GF... ?
    public void drawWaveDangers(int surfableWaveIndex) {
        DiaWave surfWave = findSurfableWave(surfableWaveIndex); 
        if (surfWave == null) {
            return;
        }
        
        double maxEscapeAngle = Math.asin(8.0 / surfWave.bulletSpeed);
        
        double[] gfDangers = new double[31];
        double minDanger = Double.POSITIVE_INFINITY;
        
        for (int x = 0; x <= 30; x++) {
            double gf = (x - 15) / 15.0;
            double bearingOffset = surfWave.orbitDirection * 
                (gf * maxEscapeAngle);
            Point2D.Double dangerLocation = DiaUtils.project(
                surfWave.sourceLocation, surfWave.absBearing + bearingOffset,
                surfWave.distanceTraveled(_currentTime));
            gfDangers[x] = 
                getDangerScore(surfWave, dangerLocation, surfableWaveIndex);

            if (gfDangers[x] < minDanger) {
                minDanger = gfDangers[x];
            }
        }
        
        double avg = DiaUtils.average(gfDangers);
        double stDev = DiaUtils.standardDeviation(gfDangers);
        
        for (int x = 0; x <= 30; x++) {
            double gf = (x - 15) / 15.0;
            double bearingOffset = surfWave.orbitDirection * 
                (gf * maxEscapeAngle);
            Point2D.Double drawLocation = DiaUtils.project(
                surfWave.sourceLocation, surfWave.absBearing + bearingOffset,
                surfWave.distanceTraveled(_currentTime) - 12);

            _renderables.add(RoboGraphic.drawCircleFilled(drawLocation, 
                 riskColor(gfDangers[x] - minDanger, avg - minDanger, 
                     stDev, false, 1), 2));
        }
    }
    
    public static Color riskColor(double risk, double avg,
        double stDev, boolean safestYellow, double maxStDev) {
            
        if (risk < .0000001 && safestYellow) {
            return Color.yellow;
        }
        
        return new Color(
            (int)DiaUtils.limit(0, 255 * (risk - (avg - maxStDev*stDev)) 
                    / (2*maxStDev*stDev), 255), 
                0,
            (int)DiaUtils.limit(0, 255 * ((avg + maxStDev*stDev) - risk) 
                    / (2*maxStDev*stDev), 255));
    }

    public void paintOn() {
        _painting = ENABLE_DEBUGGING_GRAPHICS;
    }
    
    public void paintOff() {
        _renderables.clear();
        _painting = false;
    }
    
    public boolean is1v1() {
        return (_enemiesAlive <= 1);
    }
    
    public EnemyDataMove duelOpponent() {
        return _enemies.get(_opponentName);
    }
    
    protected static final double[]  WEIGHTS_SIMPLE = 
        new double[]{1, 0, 1, 1, 0, 0, 0, 0};
    protected static final double[]  WEIGHTS_NORMAL = 
        new double[]{4, 4, 0, 0, 4, 1, 3, 3};

    // Produces a data point from one of our enemy waves. This defines
    // how we evaluate dangers.
    public double[] dataPointFromWave(DiaWave w) {
        double[] point = new double[]{
            (Math.min(1000, w.targetDistance) / 1000),
            (Math.abs(w.lateralVelocity()) / 8),
            (Math.abs(w.targetVelocity) / 8),
            (w.targetRelativeHeading / Math.PI),
            (Math.min(1.0, w.targetWallDistance) / 1.0),
            (Math.min(1.0, w.targetRevWallDistance) / 1.0),
            (Math.min(1, w.targetVchangeTime / (w.targetDistance/w.bulletSpeed))),
            (w.targetDl8t / 64)
            };
        
        return point;
    }

}

interface DistanceController {
    public double attackAngle(double currentDistance, double desiredDistance);
}

class SubtleDistancing implements DistanceController {
    public double attackAngle(double currentDistance, double desiredDistance) {
        
        double distanceOffset = currentDistance - desiredDistance;
        double attackAngle = 
            DiaUtils.square(distanceOffset / desiredDistance) 
                * DiaUtils.nonZeroSign(distanceOffset);
        
        return attackAngle;
   }
}
 
class AggressiveDistancing implements DistanceController {
    public double attackAngle(double currentDistance, double desiredDistance) {
        
        double distanceOffset = currentDistance - desiredDistance;
        double attackAngle = 
            DiaUtils.square(distanceOffset / desiredDistance) 
                * DiaUtils.nonZeroSign(distanceOffset) * 3;
        
        return attackAngle;
   }
}

abstract class MovementChoice implements java.lang.Comparable<MovementChoice> {
    public double lastDanger = Double.POSITIVE_INFINITY;
    
    abstract public int getMovementOption();
    
    public int compareTo(MovementChoice m2) {
        if (lastDanger < m2.lastDanger) { return -1; }
        else if (lastDanger > m2.lastDanger) { return 1; }
        else { return 0; }
    }
}

class MovementCounterClockwise extends MovementChoice {
    public int getMovementOption() {
        return DiamondWhoosh.COUNTERCLOCKWISE_OPTION;
    }
}

class MovementStop extends MovementChoice {
    public int getMovementOption() {
        return DiamondWhoosh.STOP_OPTION;
    }
}

class MovementClockwise extends MovementChoice {
    public int getMovementOption() {
        return DiamondWhoosh.CLOCKWISE_OPTION;
    }
}
