package voidious.move;

import java.awt.geom.Point2D;
import java.util.HashMap;
import java.util.Iterator;
import java.util.ArrayList;
import java.util.LinkedList;
import robocode.util.Utils;
import voidious.utils.DiaUtils;
import voidious.utils.KdBucketTree;

/*
 * General info about the enemy. We just keep one of these around for 
 * each enemy bot the movement is aware of.
 * 
 * Code is open source, released under the RoboWiki Public Code License:
 * http://robowiki.net/?RWPCL
 */

public class EnemyDataMove {
    public String botName;
    public double distance;
    public double energy;
    public Point2D.Double location;
    public long lastScanTime;
    public ArrayList<Point2D.Double> pastLocations;
    public double heading;
    public double absBearing;
    public double damageTaken;
    public double damageGiven;
    boolean alive;
    public double lastBulletPower;
    public double totalBulletPower;
    public long totalTimesHit;
    public long lastTimeHit;
    public long lastTimeClosest;
    protected HashMap<String, Double> _botDistancesSq;
    public KdBucketTree scanTree;
    public KdBucketTree recentScanTree;
    public KdBucketTree flattenerTree;
    public HashMap<double[], Double> guessFactors;
    public HashMap<double[], Double> flatFactors;
    public LinkedList<double[]> recentDataPoints;
    public boolean flattenerEnabled;
    public boolean avoidBeingTargeted;
    public boolean stayPerpendicular;
    
    public double weighted1v1ShotsFired;
    public double weighted1v1ShotsHit;
    public double weighted1v1ShotsFiredThisRound;
    public double weighted1v1ShotsHitThisRound;
    public double lastRoundNormalized1v1HitPercentage;

    public EnemyDataMove(String name, double dist, double en, Point2D.Double l, 
        double h, double ab, long lst) {
        
        botName = name;
        distance = dist;
        energy = en;
        location = l;
        heading = h;
        absBearing = ab;
        damageTaken = 0;
        damageGiven = 0;
        alive = true;
        lastBulletPower = 3.0;
        totalBulletPower = 0;
        totalTimesHit = 0;
        lastScanTime = lst;
        weighted1v1ShotsFired = 0;
        weighted1v1ShotsHit = 0;
        flattenerEnabled = false;
        avoidBeingTargeted = false;
        stayPerpendicular = false;
        
        _botDistancesSq = new HashMap<String, Double>();
        scanTree = new KdBucketTree();
        recentScanTree = new KdBucketTree();
        flattenerTree = new KdBucketTree();
        guessFactors = new HashMap<double[], Double>();
        flatFactors = new HashMap<double[], Double>();
        recentDataPoints = new LinkedList<double[]>();
        pastLocations = new ArrayList<Point2D.Double>();
    }
    
    public double avgBulletPower() {
        if (totalBulletPower == 0) { return 3.0; }
        
        return totalBulletPower / totalTimesHit;
    }
    
    public void setBotDistanceSq(String name, double distance) {
        _botDistancesSq.put(name, distance);
    }
    
    public void clearDistancesSq() {
        _botDistancesSq.clear();
    }
    
    public void removeDistanceSq(String botName) {
        _botDistancesSq.remove(botName);
    }
    
    public void resetRecentScanTree() {
        recentScanTree = new KdBucketTree();
    }
    
    public double minDistanceSq() {
        double min = Double.POSITIVE_INFINITY;
        
        Iterator<Double> distIterator = _botDistancesSq.values().iterator();
        while (distIterator.hasNext()) {
            double dist = distIterator.next();
            if (dist < min) {
                min = dist;
            }
        }
        
        return min;
    }
    
    public int botsCloser(double distanceSq) {
        int bc = 0;
        
        Iterator<Double> distIterator = _botDistancesSq.values().iterator();
        while (distIterator.hasNext()) {
            double dist = distIterator.next();
            if (dist < distanceSq) {
                bc++;
            }
        }
        
        return bc;
    }
    
    public Point2D.Double getPastLocation(int x) {
        return pastLocations.get(pastLocations.size() - 1 - x);
    }
}
